"""
Mosaik API for simulations written in Python.

"""

from __future__ import annotations

import abc
import asyncio
import inspect
import re
import socket
import sys
from importlib import metadata
from typing import Any, Dict, List, Optional, Tuple, Union

import docopt
from loguru import logger

from .connection import Channel, RemoteException, single_channel
from .types import (
    Attr,
    CreateResult,
    CreateResultChild,
    EntityGraph,
    EntitySpec,
    FullId,
    InputData,
    Meta,
    ModelName,
    OutputData,
    OutputRequest,
    SimId,
    Time,
)

__version__ = metadata.version("mosaik-api-v3")
__api_version__ = __version__
__all__ = [
    "__version__",
    "__api_version__",
    "Simulator",
    "start_simulation",
    "MosaikProxy",
    "Attr",
    "CreateResult",
    "CreateResultChild",
    "EntityGraph",
    "EntitySpec",
    "FullId",
    "InputData",
    "OutputRequest",
    "OutputData",
    "Meta",
    "ModelName",
    "SimId",
    "Time",
]


_HELP = """%(desc)s

Usage:
    %(prog)s [options] HOST:PORT

Options:
    HOST:PORT   Connect to this address
    -l LEVEL, --log-level LEVEL
                Log level for simulator (%(levels)s)
    -r, --remote
                Simulator is to be started on a machine remote from mosaik
    -t TIME, --timeout TIME
                Timeout in seconds for mosaik handshake [default: 60]
%(extra_opts)s
"""
_LOG_LEVELS = {
    "trace": "TRACE",
    "debug": "DEBUG",
    "info": "INFO",
    "warning": "WARNING",
    "error": "ERROR",
    "critical": "CRITICAL",
}


logger.disable(__name__)


# NOTE: We don't use an ABC here, because the effort of making it py2 AND py3
# compatible (-> meta classes) outweighs the benefits.
class Simulator(object):
    """This is the base class that you need to inherit from and implement the
    API calls."""

    meta: Meta
    """Meta data describing the simulator (the same that is returned by
    :meth:`init()`).

    ::

        {
            'api_version': 'x.y',
            'type': 'time-based'|'event-based'|'hybrid',
            'models': {
                'ModelName': {
                    'public': True|False,
                    'params': ['param_1', ...],
                    'attrs': ['attr_1', ...],
                    'any_inputs': True|False,
                    'trigger': ['attr_1', ...],
                    'non-persistent': ['attr_2', ...],
                },
                ...
            },
            'extra_methods': [
                'do_cool_stuff',
                'set_static_data'
            ]
        }

    The *api_version* is a string that defines which version of the mosaik API
    the simulator implements.  Since mosaik API version 2.3, the simulator's
    `major version <http://semver.org/>`_ ("x", in the snippet above) has to be
    equal to mosaik's.  Mosaik will cancel the simulation if a version mismatch
    occurs.

    The *type* defines how the simulator is advanced through time and whether
    its attributes are persistent in time or transient.

    *models* is a dictionary describing the models provided by this simulator.
    The entry *public* determines whether a model can be instantiated by a user
    (``True``) or if it is a sub-model that cannot be created directly
    (``False``). *params* is a list of parameter names that can be passed to
    the model when creating it. *attrs* is a list of attribute names that can
    be accessed (reading or writing).  If the optional *any_inputs* flag is set
    to ``true``, any attributes can be connected to the model, even if they are
    not *attrs*. This may, for example, be useful for databases that don't know
    in advance which attributes of an entity they'll receive. *trigger* is a
    list of attribute names that cause the simulator to be stepped when
    another simulator provides output which is connected to one of those.


    *extra_methods* is an optional list of methods that a simulator provides in
    addition to the standard API calls (``init()``, ``create()`` and so on).
    These methods can be called while the scenario is being created and can be
    used for operations that don't really belong into ``init()`` or
    ``create()``.

    """

    mosaik: MosaikProxy = None  # type: ignore # Will be set by "start_simulation()"
    # TODO: Maybe introduce abstract MosaikProxy?
    """An RPC proxy to mosaik."""

    def __init__(self, meta):
        self.meta = {
            "api_version": __api_version__,
            "models": {},
            "extra_methods": [],
            "type": "time-based",
        }
        self.meta.update(meta)

    def init(self, sid: SimId, time_resolution: float = 1.0, **sim_params) -> Meta:
        """Initialize the simulator with the ID *sid* and pass the
         *time_resolution* and additional parameters *(sim_params)*
         sent by mosaik. Return the meta data :attr:`meta`.

        If your simulator has no *sim_params*, you don't need to override this
        method.

        """
        return self.meta

    def create(self, num: int, model: ModelName, **model_params) -> List[CreateResult]:
        """Create *num* instances of *model* using the provided *model_params*.

        *num* is an integer for the number of model instances to create.

        *model* needs to be a public entry in the simulator's
        ``meta['models']``.

        *model_params* is a dictionary mapping parameters (from
        ``meta['models'][model]['params']``) to their values.

        Return a (nested) list of dictionaries describing the created model
        instances (entities). The root list must contain exactly *num*
        elements. The number of objects in sub-lists is not constrained::

            [
                {
                    'eid': 'eid_1',
                    'type': 'model_name',
                    'rel': ['eid_2', ...],
                    'children': [
                        {'eid': 'child_1', 'type': 'child'},
                        ...
                    ],
                },
                ...
            ]

        The entity ID (*eid*) of an object must be unique within a simulator
        instance. For entities in the root list, *type* must be the same as the
        *model* parameter. The type for objects in sub-lists may be anything
        that can be found in ``meta['models']``. *rel* is an optional list of
        related entities; "related" means that two entities are somehow connect
        within the simulator, either logically or via a real data-flow (e.g.,
        grid nodes are related to their adjacent branches). The *children*
        entry is optional and may contain a sub-list of entities.

        """
        raise NotImplementedError

    def setup_done(self) -> None:
        """Callback that indicates that the scenario setup is done and the
        actual simulation is about to start.

        At this point, all entities and all connections between them are know
        but no simulator has been stepped yet.

        Implementing this method is optional.

        *Added in mosaik API version 2.3*

        """

    def step(self, time: Time, inputs: InputData, max_advance: Time) -> Optional[Time]:
        """Perform the next simulation step from time *time* using input values
        from *inputs* and return the new simulation time (the time at which
        ``step()`` should be called again).

        *time* and the time returned are integers. Their unit is arbitrary,
        e.g. *seconds* (from simulation start), but has to be consistent among
        all simulators used in a simulation.

        *inputs* is a dict of dicts mapping entity IDs to attributes and
        dicts of values (each simulator has to decide on its own how to reduce
        the values (e.g., as its sum, average or maximum)::

            {
                'dest_eid': {
                    'attr': {'src_fullid': val, ...},
                    ...
                },
                ...
            }

        *max_advance* tells the simulator how far it can advance its time
        without risking any causality error, i.e. it is guaranteed that no
        external step will be triggered before max_advance + 1, unless the
        simulator activates an output loop earlier than that. For time-based
        simulators (or hybrid ones without any triggering input) *max_advance*
        is always equal to the end of the simulation (*until*).
        """
        raise NotImplementedError

    def get_data(self, outputs: OutputRequest) -> OutputData:
        """Return the data for the requested attributes in *outputs*

        *outputs* is a dict mapping entity IDs to lists of attribute names
        whose values are requested::

            {
                'eid_1': ['attr_1', 'attr_2', ...],
                ...
            }

        The return value needs to be a dict of dicts mapping entity IDs and
        attribute names to their values::

            {
                'eid_1: {
                    'attr_1': 'val_1',
                    'attr_2': 'val_2',
                    ...
                },
                ...
                'time': output_time (for event-based sims, optional)
            }

        Time-based simulators have set an entry for all requested attributes,
        whereas for event-based and hybrid simulators this is optional (e.g.
        if there's no new event).
        Event-based and hybrid simulators can optionally set a timing of their
        non-persistent output attributes via a *time* entry, which is valid
        for all given (non-persistent) attributes. If not given, it defaults
        to the current time of the step. Thus only one output time is possible
        per step. For further output times the simulator has to schedule
        another self-step (via the step's return value).

        """
        raise NotImplementedError

    def configure(self, args):
        """This method can be overridden to configure the simulation with the
        command line *args* as created by `docopt <http://docopt.org/>`_.

        *backend* and *env* are the *simpy.io* backend and environment used
        for networking. You can use them to start extra processes (e.g., a
        web server).

        The default implementation simply ignores them.

        """
        pass

    def event_setter(self):
        """This method can be overridden to allow the simulator to
        asynchronously set events for itself at time step event_time via the
        set_event api call::
            yield self.mosaik.set_event(event_time)

        *env* is the *simpy.io* environment.

        The default implementation does nothing by just yielding a triggered
        event.

        """
        pass

    def finalize(self):
        """This method can be overridden to do some clean-up operations after
        the simulation finished (e.g., shutting down external processes).

        """
        pass


class MosaikProxy(abc.ABC):
    @abc.abstractmethod
    async def get_progress(self) -> float:
        pass

    @abc.abstractmethod
    async def get_related_entities(
        self, entities: Union[None, FullId, List[FullId]]
    ) -> Union[
        EntityGraph, Dict[FullId, EntitySpec], Dict[FullId, Dict[FullId, EntitySpec]]
    ]:
        pass

    @abc.abstractmethod
    async def get_data(
        self, attrs: Dict[FullId, Attr]
    ) -> Dict[FullId, Dict[Attr, Any]]:
        pass

    @abc.abstractmethod
    async def set_data(self, data: Dict[FullId, Dict[FullId, Dict[Attr, Any]]]) -> None:
        pass

    @abc.abstractmethod
    async def set_event(self, event_time: Time) -> None:
        pass


class RemoteMosaikProxy(MosaikProxy):
    def __init__(self, channel: Channel):
        self._channel = channel

    async def get_progress(self) -> float:
        return await self._channel.send(["get_progress", [], {}])

    async def get_related_entities(
        self, entities: Union[None, FullId, List[FullId]]
    ) -> Union[
        EntityGraph, Dict[FullId, EntitySpec], Dict[FullId, Dict[FullId, EntitySpec]]
    ]:
        return await self._channel.send(["get_related_entities", [entities], {}])

    async def get_data(
        self, attrs: Dict[FullId, Attr]
    ) -> Dict[FullId, Dict[Attr, Any]]:
        return await self._channel.send(["get_data", [attrs], {}])

    async def set_data(self, data: Dict[FullId, Dict[FullId, Dict[Attr, Any]]]) -> None:
        return await self._channel.send(["set_data", [data], {}])

    async def set_event(self, event_time: Time) -> None:
        return await self._channel.send(["set_event", [event_time], {}])


def start_simulation(
    simulator: Simulator,
    description: str = "",
    extra_options=None,
    configure_logging: bool = True,
) -> int:
    """Start the simulation process for ``simulator``.

    *simulator* is the instance of your API implementation (see
        :class:`Simulator`).

    *description* may override the default description printed with the help on
    the command line.

    *extra_option* may be a list of options for `docopt <http://docopt.org/>`_
    (example: ``['-e, --example     Enable example mode']``). Commandline
    arguments are passed to :meth:`Simulator.configure()` so that your API
    implementation can handle them.

    *configure_logging* determines whether the API sets up the loguru
    logger. If the user specifies the log level using the --log-level
    command line flag, logging will always be configured.
    """
    return asyncio.run(
        start_simulation_async(
            simulator, description, extra_options, configure_logging=configure_logging
        )
    )


async def run_as_client(
    simulator: Simulator,
    host: str,
    port: int,
):
    """Run the simulator ``simulator`` as a *client* (meaning that it
    tries to connect to mosaik). This corresponds to mosaik's *cmd*
    start method.

    ``host`` and ``port`` should be the address of the mosaik instance
    to which to connect. When starting a simulator from a mosaik
    scenario, this address is inserted into the command string as
    ``%(addr)s``.
    """
    sim_name = simulator.__class__.__name__

    # Check if simulator has implemented *time_resolution* and *max_advance*:
    api_compliant = check_api_compliance(simulator)

    try:
        logger.info(f"Simulator {sim_name} tries to connect to mosaik ...")
        reader, writer = await asyncio.open_connection(host, port)
        channel = Channel(reader, writer)
    except ConnectionRefusedError:
        logger.error("Could not connect to mosaik")
        raise
    except OSError as e:
        if e.errno == 10057:  # Connection refused under Windows
            logger.error("Could not connect to mosaik")
        raise

    try:
        await run_simulator(channel, simulator, api_compliant=api_compliant)
    finally:
        await channel.close()


async def run_as_server(
    simulator: Simulator,
    host: str,
    port: int,
    *,
    start_timeout: int = 60,
):
    """Run the simulator ``simulator`` as a server (meaning that it
    waits for mosaik to establish a connection). This corresponds to
    mosaik's *connect* start method.

    ``host`` and ``port`` are the host and port to which the server will
    attempt to bind. ``start_timeout`` specifies how long to wait for
    mosaik to make the connection before aborting the program.
    """
    sim_name = simulator.__class__.__name__

    # Check if simulator has implemented *time_resolution* and *max_advance*:
    api_compliant = check_api_compliance(simulator)

    try:
        logger.info(f"Simulator {sim_name} awaits connection from mosaik ...")
        channel = await asyncio.wait_for(single_channel(host, port), start_timeout)
    except asyncio.TimeoutError:
        raise RuntimeError("Connection from mosaik not received in time")

    try:
        await run_simulator(channel, simulator, api_compliant=api_compliant)
    finally:
        await channel.close()


async def run_simulator(channel: Channel, simulator: Simulator, *, api_compliant: bool):
    """Run the Simulator ``simulator`` on the Channel ``channel``.

    Usually, you will use :func:`run_as_client` or :func:`run_as_server`
    to have the channel be created for you.

    ``api_compliant=False`` allows the use of older mosaik Simulators
    that implement version 2 of the mosaik simulator API unchanged, but
    we recommend that you update your implementation to API version 3,
    instead.
    """
    try:
        simulator.mosaik = RemoteMosaikProxy(channel)
        await init(channel, simulator, api_compliant=api_compliant)
        _ = asyncio.create_task(get_wrapper(simulator.event_setter)())
        await run(channel, simulator)
    finally:
        _ = await get_wrapper(simulator.finalize)()


async def start_simulation_async(
    simulator: Simulator,
    description: str = "",
    extra_options: List[str] | None = None,
    configure_logging: bool = False,
) -> int:
    """Run the Simulator ``simulator`` with settings read from the
    command-line arguments. This is a convenience method to quickly
    create a usable command-line interface.

    If you want more control (for example, integrating your own
    command-parsing), :func:`run_as_client` and :func:`run_as_server`
    are better alternatives.

    ``description`` will be shown as the default message when running
    the simulator program. ``extra_options`` will allow you to specify
    extra command-line arguments that will be passed to your simulator's
    :meth:`~Simulator.configure` method.

    ``configure_logging`` specifies whether logging will be set up.
    If left ``False``, you can enable loguru output for the API by
    calling ``logger.enable("mosaik_api_v3")`` and otherwise configuring
    loguru however you want.
    """
    OK, ERR = 0, 1

    args = _parse_args(
        description or "Start the simulation service.",
        extra_options or [],
    )

    log_level = args.get("--log-level")
    if configure_logging or log_level:
        logger.enable(__name__)
        logger.remove()
        logger.add(sys.stderr, level=_LOG_LEVELS.get(log_level, "INFO"))

    simulator.configure(args)

    try:
        host, port = _parse_addr(args["HOST:PORT"])
        remote_flag = args["--remote"] if "--remote" in args.keys() else False
    except ValueError as exc:
        # These are (should be) user errors (as opposed to programmer
        # errors), so we print a more friendly error message
        print(f"Invalid command-line argument: {exc}", file=sys.stderr)
        return ERR

    try:
        if remote_flag:
            start_timeout = int(args["--timeout"])
            await run_as_server(simulator, host, port, start_timeout=start_timeout)
        else:
            await run_as_client(simulator, host, port)
    except (ConnectionError, KeyboardInterrupt):
        pass  # Exit silently.

    return OK


def check_api_compliance(simulator):
    """Checks for compliance with API 3:
    i.e. if meta contains 'type' and if the new parameters,
    namely *time_resolution* for the init method and *max_advance* for step
    are implemented.

    """
    compliant = True
    sim_name = simulator.__class__.__name__

    for func_name, param in [("init", "time_resolution"), ("step", "max_advance")]:
        func = getattr(simulator, func_name)
        signature = inspect.signature(func)
        for isig in signature.parameters.values():
            if isig.name == param or str(isig.kind) == "VAR_KEYWORD":
                break
        else:
            compliant = False
            print(
                "DEPRECATION WARNING: '%s' is not implemented as argument "
                "of %s's %s function. This might cause an error in "
                "future API versions." % (param, sim_name, func_name)
            )
    return compliant


async def init(channel: Channel, sim: Simulator, *, api_compliant: bool) -> None:
    init_func = get_wrapper(sim.init)
    request = await channel.next_request()
    func, args, kwargs = request.content
    logger.debug("Calling %s(*%s, **%s)" % (func, args, kwargs))
    assert func == "init"
    sim.time_resolution = kwargs["time_resolution"]
    if not api_compliant:
        kwargs.pop("time_resolution")
    ret = await init_func(*args, **kwargs)
    await request.set_result(ret)


async def run(channel: Channel, sim):
    """Main simulator process. Send a greeting message to mosaik and wait
    for requests to step the simulation, get data or whatever.

    *channel* is a :class:`simpy.io.message.Message` instance.

    *sim* is the instance of an :class:`Simulator` implementation.

    """
    funcs = {
        "init": sim.init,
        "create": sim.create,
        "setup_done": sim.setup_done,
        "step": sim.step,
        "get_data": sim.get_data,
    }
    extra_funcs = {
        name: getattr(sim, name) for name in sim.meta.get("extra_methods", [])
    }
    funcs.update(extra_funcs)
    for name, func in funcs.items():
        funcs[name] = get_wrapper(func)

    logger.debug("Entering event loop ...")
    while True:
        try:
            request = await channel.next_request()
        except asyncio.IncompleteReadError:
            logger.info("mosaik closed its connection unexpectedly")
            break
        func, args, kwargs = request.content
        logger.debug("Calling %s(*%s, **%s)" % (func, args, kwargs))
        if func == "stop":
            break

        try:
            func = funcs[func]
            ret = await func(*args, **kwargs)
            await request.set_result(ret)
        except Exception as e:
            await request.set_exception(e)


def get_wrapper(func):
    if inspect.isgeneratorfunction(func):

        async def wrapper(*args, **kwargs):
            gen = func(*args, **kwargs)
            try:
                # Requests from the simulator to mosaik are encoded as
                # coroutines that the simulator yields (not awaits!).
                # So we will have to do the awaiting.
                # (This is due to the way things worked in simpy to
                # avoid breaking existing simulators.)
                request = next(gen)
                while True:
                    try:
                        # Awaiting the request gives the result or
                        # raises an exception.
                        remote_result = await request
                    except RemoteException as remote_exception:
                        request = gen.throw(remote_exception)
                    else:
                        request = gen.send(remote_result)
            except StopIteration as stop:
                return stop.value
    else:

        async def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

    wrapper.__name__ = func.__name__
    return wrapper


def _parse_args(desc, extra_options):
    """Fill-in the values into :data:`_HELP` and parse and return the arguments
    using :func:`~docopt.docopt()`.

    """
    msg = _HELP % {
        "desc": desc,
        "prog": sys.argv[0],
        "levels": ", ".join(_LOG_LEVELS.keys()),
        "extra_opts": "\n".join("    %s" % opt for opt in extra_options),
    }
    args = docopt.docopt(msg)
    args["--timeout"] = args.get("--timeout", 60)
    return args


def _parse_addr(addr: str) -> Tuple[str, int]:
    """Parse ``addr`` and returns a ``('host', port)`` tuple.

    If the host does not look like an IP(v4) address, resolve its name
    to an IP address.

    Raise a :exc:`ValueError` if resolving the hostname fails or if the
    address contains no host or port.

    """
    try:
        host, port = addr.strip().split(":")
    except ValueError:
        raise ValueError(f"address '{addr}' does not match expected format HOST:PORT")

    try:
        # Resolve hostname if it doesn't look like an IP address
        if not re.match(r"^(\d{1,3}\.){3}\d{1,3}$", host):
            host = socket.gethostbyname(host)
    except (IOError, OSError):
        raise ValueError(f"could not resolve host '{host}'")

    return (host, int(port))



"""
The example simulator contains two models:

- Model A produces continuously increasing integers.
- Model B consumes a value and yields the last value consumed.

"""


def model_a(i):
    while True:
        yield i
        i += 1


def model_b(val):
    while True:
        new_val = yield val
        if new_val is not None:
            val = new_val


class Simulator(object):
    models = {
        "A": model_a,
        "B": model_b,
    }

    def __init__(self, model, num_inst, init_val):
        self.model = model
        self.instances = []
        self.results = []

        for i in range(num_inst):
            inst = Simulator.models[model](init_val)
            self.results.append(next(inst))  # Initialize generator
            self.instances.append(inst)

    def step(self, inputs):
        self.results = [inst.send(input) for inst, input in zip(self.instances, inputs)]



# This file exists because pyproject.toml files offer no built-in way of
# combining several files to form the project description on PyPI,
# but we would like to have the README.rst, CHANGES.rst and AUTHORS.rst
# there. Luckily, we can fill the readme field in the project table
# dynamically using a Hatch plugin.

from hatchling.metadata.plugin.interface import MetadataHookInterface


class DescriptionDataHook(MetadataHookInterface):
    def update(self, metadata):
        metadata["readme"] = {
            "content-type": "text/x-rst",
            "text": "\n\n".join(
                open(f, encoding="utf-8").read()
                for f in ["README.rst", "CHANGES.txt", "AUTHORS.txt"]
            ),
        }



"""
Mosaik interface for the example simulator.

It more complex than it needs to be to be more flexible and show off various
features of the mosaik API.

"""

import logging

import mosaik_api_v3
from example_sim import simulator

logger = logging.getLogger("example_sim")


example_sim_meta = {
    "type": "time-based",
    "models": {
        "A": {
            "public": True,
            "params": ["init_val"],
            "attrs": ["val_out", "dummy_out"],
        },
        "B": {
            "public": True,
            "params": ["init_val"],
            "attrs": ["val_in", "val_out", "dummy_in"],
        },
        "C": {  # Doesn't acutally exist. Used for testing.
            "public": False,
            "attrs": [],
        },
    },
    "extra_methods": [
        "example_method",
    ],
}


class ExampleSim(mosaik_api_v3.Simulator):
    def __init__(self):
        super(ExampleSim, self).__init__(example_sim_meta)
        self.step_size = None
        self.simulators = []
        self.value = None  # May be set in example_method()

    def init(self, sid, time_resolution, step_size=1):
        self.sid = sid
        self.step_size = step_size
        return self.meta

    def create(self, num, model, init_val):
        sim_id = len(self.simulators)
        sim = simulator.Simulator(model, num, init_val)
        self.simulators.append(sim)
        return [
            {"eid": "%s.%s" % (sim_id, eid), "type": model, "rel": []}
            for eid, inst in enumerate(sim.instances)
        ]

    def step(self, time, inputs, max_advance):
        for sid, sim in enumerate(self.simulators):
            sim_inputs = [None for i in sim.instances]
            for i, _ in enumerate(sim_inputs):
                eid = "%s.%s" % (sid, i)
                if eid in inputs:
                    sim_inputs[i] = sum(inputs[eid]["val_in"].values())

            for i in range(self.step_size):
                sim.step(sim_inputs)

        return time + self.step_size

    def get_data(self, outputs):
        data = {}
        for eid, attrs in outputs.items():
            sid, idx = map(int, eid.split("."))
            data[eid] = {}
            for attr in attrs:
                if attr != "val_out":
                    continue
                data[eid][attr] = self.simulators[sid].results[idx]
        return data

    def example_method(self, value):
        self.value = value
        return value


def main():
    return mosaik_api_v3.start_simulation(ExampleSim())


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any, Dict, Iterable, List, Tuple

from typing_extensions import Literal, TypeAlias, TypedDict

Time: TypeAlias = int
"""Time is represented as the number of simulation steps since the
simulation started. One step respresents `time_resolution` seconds."""
Attr: TypeAlias = str
"""An attribute name"""
ModelName: TypeAlias = str
"""The name of a model."""

ModelDescriptionOptionals = TypedDict(
    "ModelDescriptionOptionals",
    {
        "any_inputs": bool,
        "trigger": Iterable[Attr],
        "non-trigger": Iterable[Attr],
        "persistent": Iterable[Attr],
        "non-persistent": Iterable[Attr],
    },
    total=False,
)


class ModelDescription(ModelDescriptionOptionals):
    """Description of a single model in `Meta`"""

    public: bool
    """Whether the model can be created directly."""
    params: List[str]
    """The parameters given during creating of this model."""
    attrs: List[Attr]
    """The input and output attributes of this model."""


class MetaOptionals(TypedDict, total=False):
    extra_methods: List[str]
    """The names of the extra methods this simulator supports."""


class Meta(MetaOptionals):
    """The meta-data for a simulator."""

    api_version: Literal["3.0"]
    """The API version that this simulator supports in the format "major.minor"."""
    type: Literal["time-based", "event-based", "hybrid"]
    """The simulator's stepping type."""
    models: Dict[ModelName, ModelDescription]
    """The descriptions of this simulator's models."""


SimId: TypeAlias = str
"""A simulator ID"""
EntityId: TypeAlias = str
"""An entity ID"""
FullId: TypeAlias = str
"""A full ID of the form "sim_id.entity_id" """
InputData: TypeAlias = Dict[EntityId, Dict[Attr, Dict[FullId, Any]]]
"""The format of input data for simulator's step methods."""
OutputRequest: TypeAlias = Dict[EntityId, List[Attr]]
"""The requested outputs for get_data. For each entity where data is
needed, the required attributes are listed."""
OutputData: TypeAlias = Dict[EntityId, Dict[Attr, Any]]
"""The format of output data as return by ``get_data``"""


class CreateResultOptionals(TypedDict, total=False):
    rel: List[EntityId]
    """The entity IDs of the entities of this simulator that are
    related to this entity."""
    children: List[CreateResult]
    """The child entities of this entity."""
    extra_info: Any
    """Any additional information about the entity that the simulator
    wants to pass back to the scenario.
    """


class CreateResult(CreateResultOptionals):
    """The type for elements of the list returned by `create` calls in
    the mosaik API."""

    eid: EntityId
    """The entity ID of this entity."""
    type: ModelName
    """The model name (as given in the simulator's meta) of this entity.
    """


CreateResultChild: TypeAlias = CreateResult


class EntitySpec(TypedDict):
    type: ModelName


class EntityGraph(TypedDict):
    nodes: Dict[FullId, EntitySpec]
    edges: List[Tuple[FullId, FullId, Dict]]



import datetime as dt


class Converter:
    _start_time: dt.datetime
    _time_resolution: dt.timedelta

    def __init__(self, start_date: str, time_resolution: float):
        self._start_time = dt.datetime.fromisoformat(start_date)
        self._time_resolution = dt.timedelta(seconds=time_resolution)

    def datetime_from_step(self, step: int) -> dt.datetime:
        return self._start_time + step * self._time_resolution

    def isoformat_from_step(self, step: int) -> str:
        return self.datetime_from_step(step).isoformat()



"""
This is not really a mutli-agent system (MAS) but shows of how a MAS (or other
control strategy) can get or set data from and to other simulators alghough
the scenario didn't describe explicit data flows.

These remote calls to mosaik can also be used to implement data storage
backends or visualizations.

"""

import logging

import mosaik_api_v3

logger = logging.getLogger("example_mas")


example_sim_meta = {
    "type": "time-based",
    "models": {
        "Agent": {
            "public": True,
            "params": [],
            "attrs": [],
        },
    },
}


class ExampleMas(mosaik_api_v3.Simulator):
    def __init__(self):
        super(ExampleMas, self).__init__(example_sim_meta)
        self.sid = None
        self.step_size = None
        self.agents = []
        self.rel = None

    def init(self, sid, time_resolution, step_size=1):
        self.sid = sid
        self.step_size = step_size
        return self.meta

    def create(self, num, model):
        if model != "Agent":
            raise ValueError('Can only create "Agent" models.')

        num_agents = len(self.agents)
        agents = [
            {"eid": str(eid), "type": model, "rel": []}
            for eid in range(num_agents, num_agents + num)
        ]
        self.agents.extend(agents)
        return agents

    def step(self, time, inputs, max_advance):
        prog = yield self.mosaik.get_progress()
        print("Progress: %.1f%%" % prog)

        if time == 0:
            self.rel = yield self.mosaik.get_related_entities(
                ["%s.%s" % (self.sid, a["eid"]) for a in self.agents]
            )
            print(self.rel)

        data = yield self.mosaik.get_data(
            {eid: ["val_out"] for rels in self.rel.values() for eid in rels}
        )
        print(data)

        inputs = {}
        for a in self.agents:
            full_id = "%s.%s" % (self.sid, a["eid"])
            inputs[full_id] = {eid: {"val_in": 23} for eid in self.rel[full_id]}
        yield self.mosaik.set_data(inputs)

        return time + self.step_size

    def get_data(self, outputs):
        # We have nothing to give ...
        return {}


def main():
    return mosaik_api_v3.start_simulation(ExampleMas())



"""This module implements the basic communication protocol for mosaik."""

from __future__ import annotations

import asyncio
import itertools
import sys
import traceback
import weakref
from asyncio import StreamReader, StreamWriter
from json import JSONDecoder, JSONEncoder
from typing import Any, Dict, Iterator, Optional, Tuple, Union

REQUEST = 0
SUCCESS = 1
FAILURE = 2


_encoder = JSONEncoder()


def encode(obj: Any) -> bytes:
    """Encode an object as expected by mosaik communicaction protocol.

    The object is first converted to a JSON string. This string is
    prepended with its length as a 4-byte integer in big endian order.
    """
    # JSONEncoder encodes to string, then encode that string into bytes
    obj_bytes = _encoder.encode(obj).encode()
    return len(obj_bytes).to_bytes(4, "big") + obj_bytes


_decoder = JSONDecoder()


async def decode(reader: asyncio.StreamReader) -> Any:
    """Decode the next object from the StreamReader, in the format
    produced by `encode`.

    If the connection is closed before the message is complete, the
    asyncio.IncompleteReadError is passed on."""
    msg_length_bytes = await reader.readexactly(4)
    msg_length = int.from_bytes(msg_length_bytes, "big")
    msg_bytes = await reader.readexactly(msg_length)
    return _decoder.decode(msg_bytes.decode())


class RemoteException(Exception):
    """An exception due to a failure from a remote connected through a
    Channel."""

    remote_type: str
    """The exception type reported by the remote."""
    remote_msg: str
    """The message from the remote."""
    further_args: Tuple[str, ...]
    """Additional information from the remote, like a stack trace."""
    source: Optional[str]
    """The source of the exception."""

    def __init__(
        self,
        remote_type: str,
        remote_msg: str,
        *further_args: str,
        source: Optional[str] = None,
    ):
        super().__init__(remote_type, remote_msg, *further_args, source)
        self.remote_type = remote_type
        self.remote_msg = remote_msg
        self.further_args = further_args
        self.source = source


class Request:
    """An incoming request from a channel.

    Respond to the request by calling set_result (if successful) or
    set_exception (if unsuccessful)."""

    _channel: weakref.ref[Channel]
    _msg_id: int
    content: Any

    def __init__(self, channel: Channel, msg_id: int, content: Any):
        self._channel = weakref.ref(channel)
        self._msg_id = msg_id
        self.content = content

    async def set_result(self, result: Any):
        """Respond to the request with a successful result."""
        channel = self._channel()
        if channel:
            await channel._write_success(self._msg_id, result)

    async def set_exception(self, exception: Exception):
        """Respond to the request with a failure."""
        channel = self._channel()
        if channel:
            await channel._write_failure(
                self._msg_id,
                [
                    type(exception).__name__,
                    str(exception),
                    *traceback.format_exception(*sys.exc_info()),
                ],
            )


class EndOfRequests(Exception):
    pass


class Channel:
    """A Channel for mosaik-style request-response communication.

    To send out a request, call send and await the result.

    To react to incoming requests, get the next incoming request from
    next_request as a Request object and respond to it using the
    object's set_result and set_exception methods."""

    _msg_counter: Iterator[int]
    _outgoing_request_futures: Dict[int, asyncio.Future[Any]]
    _incoming_requests: asyncio.Queue[Union[Request, EndOfRequests]]
    _reader: asyncio.StreamReader
    _writer: asyncio.StreamWriter
    _receiver_task: asyncio.Task[None]
    _name: Optional[str]

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        name: Optional[str] = None,
    ):
        """Initialize a new channel.

        This method sets up a listener task for incoming requests and
        can therefore only be called while an asyncio loop is running.

        (reader, writer) should be a pair returned by
        asyncio.open_connection or received by an asycnio.start_server
        callback.
        """
        self._reader = reader
        self._writer = writer
        self._msg_counter = itertools.count()
        self._outgoing_request_futures = {}
        self._incoming_requests = asyncio.Queue()
        self._name = name
        try:
            self._receiver_task = asyncio.get_running_loop().create_task(
                self._receive_forever(),
                name=f"Reader task for {name}",
            )
        except RuntimeError as e:
            raise RuntimeError(
                "can only create Channel while event loop is running"
            ) from e

    async def _receive_forever(self):
        try:
            while True:
                msg_type, msg_id, content = await decode(self._reader)
                if msg_type == REQUEST:
                    await self._incoming_requests.put(Request(self, msg_id, content))
                elif msg_type == SUCCESS:
                    self._outgoing_request_futures.pop(msg_id).set_result(content)
                elif msg_type == FAILURE:
                    self._outgoing_request_futures.pop(msg_id).set_exception(
                        RemoteException(*content, source=self._name)
                    )
        except asyncio.IncompleteReadError as e:
            for future in self._outgoing_request_futures.values():
                if not future.cancelled():
                    future.set_exception(e)
            self._incoming_requests.put_nowait(EndOfRequests())

    async def send(self, value: Any) -> Any:
        """Sends a request over the channel and waits for the response.

        On a successful response, its value is returned.

        On an unsuccessful respose, the error is raised as a
        RemoteException.

        If the connection is closed while waiting for the response,
        an asyncio.IncompleteReadError is raised.
        """
        msg_id = next(self._msg_counter)
        result_future: asyncio.Future[Any] = asyncio.Future()
        self._outgoing_request_futures[msg_id] = result_future
        self._writer.write(encode([REQUEST, msg_id, value]))
        await self._writer.drain()
        return await result_future

    async def next_request(self) -> Request:
        """Get the next incoming request as a Request object.

        If the connection is closed while waiting for a request,
        an asyncio.IncompleteReadError is raised.
        """
        request_or_error = await self._incoming_requests.get()
        if isinstance(request_or_error, Request):
            return request_or_error
        else:
            raise request_or_error

    async def _write_success(self, msg_id: int, value: Any):
        """Write the given value to the channel as a successful response
        to the request msg_id.
        """
        self._writer.write(encode([SUCCESS, msg_id, value]))
        await self._writer.drain()

    async def _write_failure(self, msg_id: int, value: Any):
        """Write the given value to the channel as a failure response
        to the request msg_id."""
        self._writer.write(encode([FAILURE, msg_id, value]))
        await self._writer.drain()

    async def close(self):
        """Close the connection."""
        self._writer.close()
        await self._writer.wait_closed()
        self._receiver_task.cancel()


async def single_channel(host: str, port: int) -> Channel:
    """Runs a server that will only accept a single connection and
    returns this connection as a Channel."""
    channel_future: asyncio.Future[Channel] = asyncio.Future()

    async def connected_cb(reader: StreamReader, writer: StreamWriter):
        channel_future.set_result(Channel(reader, writer))

    server = await asyncio.start_server(connected_cb, host, port)
    try:
        return await channel_future
    finally:
        server.close()
